/***************************************************************************
                        vlmacros.h  -  description
                            -------------------
    begin                : Fri Aug 08 2002
    copyright            : (C) 2002 by Fabian Wenzel
    email                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#ifndef _VLMACROS_H_
#define _VLMACROS_H_

#include "vlconfigitem.h"
#include "vlslice.h"

#include <list>
#include <vector>
#include <typeinfo>

/***************************************************************************
 *
 * This file contains traits and preprocessor macros for easy integration
 * of vistaLab-Modules into the framework.
 *
 * Basically there are three categories of items that have to be integrated
 * for each module:
 *
 * - Pins of the module (i.e. Input/Output pins and render/log pins)
 * - Configurable items which can be modified by the user in a dialog
 * - Validator functions for the configurable items
 *
 * Integration of these items is done by...
 *
 * - Declaring the module class as VISTALABMODULE(Classname) (in header file)
 *
 * - "wrapping" the corresponding members of the module class into
 *    a preprocessor macro (in header file)
 *
 * - "configuring" the wrapped members by giving them a descriptive string
 *   in the cpp - file inside the macro-defined function BEGIN_CONFIG(...)
 *   and END_CONFIG()
 *
 *
 ***************************************************************************/

namespace vistaLab {

class module;


/***************************************************************************
 *
 * Each I/O - pin builds own functions which grab the correct data from
 * the framework for each new frame and store then after processing.
 *
 * TODO: Declaration as a struct is not necessary but the GCC-Compiler does
 * not compile vectors of function pointers due to a buggy default copy
 * constructor
 *
 ***************************************************************************/

typedef void (module::*TSliceHook)     (slice &currentSlice);
typedef bool (module::*TPostSliceHook) (slice &currentSlice);

struct STMPSliceHook{
    TSliceHook func;
};

struct STMPPostSliceHook{
    TPostSliceHook func;
};

typedef std::vector< STMPSliceHook >     TSliceFuncList;
typedef std::vector< STMPPostSliceHook > TPostSliceFuncList;


/***************************************************************************
 *
 * Each Choice builds own functions which initialize the correct data from
 * the available choices when initializing modules for processing. This
 * is necessary as the user might have configured the module between to
 * "run"s.
 *
 * TODO: Declaration as a struct is not necessary but the GCC-Compiler does
 * not compile vectors of function pointers due to a buggy default copy
 * constructor
 *
 ***************************************************************************/
typedef bool (configurableObject::*TInitHook ) ();

struct STMPInitHook{
    TInitHook func;
};

typedef std::vector< STMPInitHook > TInitFuncList;
typedef std::vector< configItem   > TConfigItems;

/***************************************************************************
 *
 * There are some new data types for configurable items which are
 * ONLY used as configuration repositories. There is not a single
 * instance of these dataTypes, but they have to be declared in order to
 * be usable with the traits below.
 *
 ***************************************************************************/

class fileName{
};

class saveFileName{
};

class multipleFileNames{
};

class customChoice{
};

template <class T> class singleChoice{
};

template <class T> class multipleChoice{
};


/***************************************************************************
 *
 * The following traits are configuration repositories which bind
 * configurable dataTypes to the corresponding "configtype" for runtime
 * identification (this is needed in CONFIGDATA below)
 *
 ***************************************************************************/
template < class T > struct configTraitsData{
    enum { configType = configItem::INVALID };
    typedef T dataType;
};

template<> struct configTraitsData< std::string > {
    enum { configType = configItem::STRING };
    typedef std::string dataType;
};

template<> struct configTraitsData< customChoice > {
    enum { configType = configItem::CUSTOMCHOICE };
    typedef std::string dataType;
};

template<> struct configTraitsData< fileName > {
    enum { configType = configItem::FILENAME };
    typedef std::string dataType;
};

template<> struct configTraitsData< saveFileName > {
    enum { configType = configItem::SAVEFILENAME };
    typedef std::string dataType;
};

template<> struct configTraitsData< bool > {
    enum { configType = configItem::BOOL };
    typedef bool dataType;
};

template<> struct configTraitsData< int > {
    enum { configType = configItem::INT };
    typedef int dataType;
};

template<> struct configTraitsData< double > {
    enum { configType = configItem::DOUBLE };
    typedef double dataType;
};

template<> struct configTraitsData< multipleFileNames > {
    enum { configType = configItem::MULTIPLEFILENAMES };
    typedef std::vector< std::string > dataType;
};


/***************************************************************************
 *
 * The following traits are configuration repositories which bind
 * configurable choices to the corresponding "configType" for runtime
 * identification (this is needed in CONFIGCHOICE below). They need
 * to have special typedefs
 *
 ***************************************************************************/
template< class T > struct configTraitsData< singleChoice<T> > {
    enum { configType = configItem::SINGLECHOICE };
    typedef T dataType;
    typedef int selectionsDataType;
    typedef std::vector< T > choicesDataType;
    static inline bool transfer(selectionsDataType &selection,
        dataType &target, choicesDataType &choices)
    {
        if(selection >= (int) choices.size())
            return false;

        target = choices[selection];
        return true;
    }

    static inline void preselect(selectionsDataType &selection,
        dataType &target, choicesDataType &choices)
    {
        selection=0;

        for(unsigned int i=0;i<choices.size();++i)
            if(choices[i] == target)
                selection = i;
    }
};

template< class T > struct configTraitsData< multipleChoice<T> > {
    enum { configType = configItem::MULTIPLECHOICE };
    typedef std::vector< T >   dataType;
    typedef std::vector< int > selectionsDataType;
    typedef std::vector< T >   choicesDataType;
    static inline bool transfer(selectionsDataType &selections,
        dataType &target, choicesDataType &choices)
    {
        target.clear();
        for(unsigned int i=0;i<selections.size();i++){
            if(selections[i] >= (int) choices.size())
                return false;
            target.push_back(choices[selections[i]]);
        }
        return true;
    }
    static inline void preselect(selectionsDataType &selections,
        dataType &target, choicesDataType &choices)
    {
        selections.clear();

        for(unsigned int i=0;i<target.size();++i)
            for(unsigned int j=0;j<choices.size();++j)
                if (choices[j] == target[i])
                    selections.push_back(j);
    }
};

/***************************************************************************
 *
 * The following traits are configuration repositories which bind
 * configurable pins to the corresponding "configType" for runtime
 * identification (this is needed in CONFIGPIN below).
 *
 ***************************************************************************/
struct configTraitsInputPin{
    enum { configType = configItem::INPUTPIN };
};

struct configTraitsOutputPin{
    enum { configType = configItem::OUTPUTPIN };
};



/***************************************************************************
 *
 * The following macro defines the interface for accessing configuration
 * stuff from the framework (for cleaning up, pre- and postprocessing and
 * initialization)
 *
 ***************************************************************************/
#define VISTALABTOPMODULE(a) \
    typedef a TModuleClass; \
public: \
    virtual bool _initConfig(bool repositoryModule);   \
    virtual vistaLab::TConfigItems *getConfigItems() ; \
private: \
    vistaLab::TSliceFuncList m_preProcessHooks;                  \
    vistaLab::TPostSliceFuncList m_postProcessHooks;             \
    vistaLab::TSliceFuncList m_cleanupHooks;                     \
    vistaLab::TConfigItems   m_configItems;                      \
    virtual vistaLab::TSliceFuncList *getPreProcessHooks();      \
    virtual vistaLab::TPostSliceFuncList *getPostProcessHooks(); \
    virtual vistaLab::TSliceFuncList *getCleanupHooks();

#define VISTALABMODULE(a) \
  public: \
  virtual module *clone() const \
{ \
    return new a (*this); \
} \
    VISTALABTOPMODULE(a)



/***************************************************************************
 *
 * The following macro is a wrapper for a configurable dataType.
 * It basically saves its type in a "typedef" for later configuration
 * (see CONFIGDATA).
 *
 ***************************************************************************/
#define PARAMETER(a,b) \
    vistaLab::configTraitsData<a>::dataType b;  \
    typedef vistaLab::configTraitsData<a> b##_configType;


/***************************************************************************
 *
 * The following macros are wrappers for a input/outputpins of a special
 * dataType. They define individual functions which transfer the
 * data from the slice to the target variable (pre##_b) and another
 * function (b##_addHook) which adds pre##_b to the preProcessHooks.
 * This cannot be done in CONFIGPIN as inputpin and outputpin have
 * to hook up different things (pre##_b vs. post##_b and cleanup_##b), so
 * b##_addHooks serves as a common interface such that CONFIGPIN can
 * handle both, input and output pins (see CONFIGPIN).
 *
 * The first version tried not to transfer dataTypes from the slice
 * to member variables via an own function but with a struct
 * containing the slot and a member-void pointer to the target
 * data but it is not possible to have a member-void pointer in C++
 *
 ***************************************************************************/
#define INPUTPIN(a,b) \
    const a *b; \
    typedef a b##_type; \
    typedef vistaLab::configTraitsInputPin  b##_configType;     \
    int slot_##b; \
    void pre_##b(vistaLab::slice &currentSlice) \
{ \
    b = (const a *) currentSlice[slot_##b]; \
} \
    void b##_addHooks() \
{ \
    vistaLab::STMPSliceHook newhook; \
    newhook.func = (vistaLab::TSliceHook) &TModuleClass::pre_##b; \
    getPreProcessHooks()->push_back(newhook); \
}

#define OUTPUTPIN(a,b) \
    a *b; \
    typedef a b##_type; \
    typedef vistaLab::configTraitsOutputPin b##_configType; \
    int slot_##b; \
    bool post_##b(vistaLab::slice &currentSlice) \
{ \
    if (!b) \
    return false; \
    currentSlice[slot_##b] = b; \
    return true; \
} \
    void cleanup_##b(vistaLab::slice &currentSlice) \
{ \
    delete (a *) currentSlice[slot_##b]; \
    currentSlice[slot_##b] = 0; \
} \
    void pre_##b(vistaLab::slice &currentSlice) \
{ \
    b = (a *) currentSlice[slot_##b]; \
} \
    void b##_addHooks() \
{ \
    vistaLab::STMPPostSliceHook newposthook; \
    vistaLab::STMPSliceHook     newhook; \
    newposthook.func = (vistaLab::TPostSliceHook) &TModuleClass::post_##b; \
    getPostProcessHooks()->push_back(newposthook); \
    newhook.func = (vistaLab::TSliceHook) &TModuleClass::cleanup_##b; \
    getCleanupHooks()->push_back(newhook); \
    newhook.func = (vistaLab::TSliceHook) &TModuleClass::pre_##b; \
    getPreProcessHooks()->push_back(newhook); \
}


/***************************************************************************
 *
 * The following macros are wrappers for a choice item (single or multiple)
 * for a specific dataType. They define functions which update the choice
 * with the chosen one of the "repository" of choices every time the modules
 * are initialized for processing. These functions can be added directly
 * so there is no need for another wrapper function (see CONFIGCHOICE).
 *
 ***************************************************************************/
#define CHOICEPARAM(a,b) \
    vistaLab::configTraitsData< a >::dataType b; \
    vistaLab::configTraitsData< a >::choicesDataType b##_choices; \
    vistaLab::configTraitsData< a >::selectionsDataType b##_selections; \
    typedef vistaLab::configTraitsData< a > b##_configType; \
    bool init_##b() \
{ \
    return vistaLab::configTraitsData< a >::transfer(b##_selections, b, b##_choices); \
}



/***************************************************************************
 *
 * The following macros define the "head" and "tail"
 * for the cpp - config - function which actually performs registration
 * of configurable items in the framework (the one which has to be defined
 * for all VISTAMODULE's). There are two versions: The first one is defined
 * for the baseclass, the second one for child classes. The only difference
 * is that the childclasses config-function first traverses up and calls
 * the parent's one (until the topobject's one is called and all inherited
 * levels have been configured). This gets done in the constructor of
 * repository.
 *
 ***************************************************************************/
#define BEGIN_CONFIG_TOPOBJECT(a) \
vistaLab::TSliceFuncList *a::getPreProcessHooks() \
{ \
    return &m_preProcessHooks; \
} \
vistaLab::TPostSliceFuncList *a::getPostProcessHooks() \
{ \
    return &m_postProcessHooks; \
} \
vistaLab::TSliceFuncList *a::getCleanupHooks() \
{ \
    return &m_cleanupHooks; \
} \
vistaLab::TConfigItems *a::getConfigItems() \
{ \
    return &m_configItems; \
} \
bool a::_initConfig(bool repositoryModule) \
{ \
    getPreProcessHooks() ->clear(); \
    getPostProcessHooks()->clear(); \
    getCleanupHooks()    ->clear(); \
    getConfigItems()     ->clear(); \
    getInitHooks()       ->clear();


#define BEGIN_CONFIG(a,b) \
BEGIN_CONFIG_TOPOBJECT(a) \
if(!b::_initConfig(repositoryModule)) \
return false;

#define END_CONFIG() return true; }



/***************************************************************************
 *
 * The following macros define individual calls for configurable items
 * (i.e. I/O - pins, data-items, choices and
 * render/log/validator - functions. They add configurable items to
 * the config-manager.
 *
 ***************************************************************************/
#define CONFIGPARAMETER(a, b) \
    if( (vistaLab::configItem::EType) a##_configType::configType == \
            vistaLab::configItem::VALIDATORFUNC   || \
        (vistaLab::configItem::EType) a##_configType::configType == \
            vistaLab::configItem::INPUTPIN        || \
        (vistaLab::configItem::EType) a##_configType::configType == \
            vistaLab::configItem::OUTPUTPIN       ) \
        return false; \
    addConfig( (vistaLab::configItem::EType) a##_configType::configType, \
        (void *) (&a), b);

#define CONFIGVAL(a) \
    addConfigFunc(vistaLab::configItem::VALIDATORFUNC, \
        (vistaLab::configItem::TMemberFunc) (vistaLab::configItem::TValidatorFunc) &TModuleClass::a);

#define CONFIGRENDERPIN(a,b) \
    addConfigFunc(vistaLab::configItem::RENDEROUTPUTPIN, \
        (vistaLab::configItem::TMemberFunc) (vistaLab::configItem::TRenderFunc) &TModuleClass::a, b);

#define CONFIGLOGPIN(a,b) \
    addConfigFunc(vistaLab::configItem::LOGOUTPUTPIN, \
        (vistaLab::configItem::TMemberFunc) (vistaLab::configItem::TLogFunc) &TModuleClass::a, b);


#define CONFIGOUTPUTPIN(a, b) \
    if ( (vistaLab::configItem::EType) a##_configType::configType != \
        vistaLab::configItem::OUTPUTPIN ) \
        return false; \
    std::vector< std::string > a##tmp; \
    a##tmp.push_back(typeid(a).name()); \
    if (repositoryModule) \
        slot_##a = -1; \
    addConfig( (vistaLab::configItem::EType) a##_configType::configType, \
        (void *) (&slot_##a),b,a##tmp); \
    a##_addHooks();

#define CONFIGINPUTPIN(a, b)  \
    if ( (vistaLab::configItem::EType) a##_configType::configType != \
        vistaLab::configItem::INPUTPIN  ) \
        return false; \
    std::vector< std::string > a##tmp; \
    a##tmp.push_back(typeid(const_cast<a##_type *>(a)).name()); \
    if (repositoryModule) \
        slot_##a = -1; \
    addConfig( (vistaLab::configItem::EType) a##_configType::configType, \
        (void *) (&slot_##a),b,a##tmp); \
    a##_addHooks();


#define CONFIGCHOICEPARAM(a,b,c,d) \
    addConfig( (vistaLab::configItem::EType) a##_configType::configType, \
        (void *) (&a##_selections), b, c); \
    a##_choices = d; \
    if (repositoryModule) \
        a##_configType::preselect(a##_selections,a,a##_choices); \
    vistaLab::STMPInitHook hook_##a; \
    hook_##a.func = (vistaLab::TInitHook) &TModuleClass::init_##a; \
    getInitHooks()->push_back(hook_##a);

#define CONFIGUNITTEST(a, b) \
    addConfigFunc(vistaLab::configItem::UNITTESTPIN, \
    (vistaLab::configItem::TMemberFunc) (vistaLab::configItem::TUnitTest) &TModuleClass::a,b);


/***************************************************************************
 *
 * The following macros are helper-macros for copy-constructors as the
 * developer does not need to know the background of all the helper
 * variables which are used here. They are necessary only for pins and
 * choices as these are the two types which have helper variables.
 * Renderfunctions (obviously) and data items do not have auxillary
 * items which have to be copied from another class.
 *
 ***************************************************************************/
#define COPYPIN(a,b) \
    slot_##a = b.slot_##a;

#define COPYCHOICE(a,b) \
    a##_choices    = b.a##_choices; \
    a##_selections = b.a##_selections;


#define EXPORTVISTALABMODULE(a) \
    extern "C" { VISTALABMODULEEXPORT vistaLab::module *getModule () { return new a(); } }

} // namespace vistaLab

#endif // #ifndef _VLMACROS_H_
