/***************************************************************************
                      vlmodule.h  -  description
                            -------------------
    begin                : Mon Feb 4 2002
    copyright            : (C) 2002 by Fabian Wenzel
    eMail                : wenzel@tu-harburg.de

***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This file is released under the the terms of the VistaLab public        *
 * License V1.5. See file COPYING for details.                             *
 *                                                                         *
 ***************************************************************************/

#ifndef _VLMODULE_H_
#define _VLMODULE_H_

#include "vlmacros.h"
#include "vldef.h"
#include "vlslice.h"
#include "vlconfigitem.h"

#include <string>
#include <vector>
#include <list>

namespace vistaLab {

/***************************************************************************/
//! The topmost class for all objects which might be configured
/*!
 *  Each module is a configurable object as well as the document, i.e.
 *  each class which might want to be configured via dialog/file
 *
 * \warning The localization - macro _() is included for all childs such
 *          that this class is not totally platform independent
 *
 * \author  Fabian Wenzel
 *
 ***************************************************************************/
class VISTALABBASEEXPORT configurableObject {
public:
    const std::string &getName() const
    {
        return m_name;
    }

    virtual ~configurableObject()
    {
    }

    virtual TConfigItems *getConfigItems()=0;

    // After configuration, the module might try to do things itself
    virtual void postConfig() {}

    // Inline - function for choice-item initialization (called when
    // initializing the module for processing)
    bool initChoices()
    {
        bool successful = true;
        TInitFuncList::iterator currentHook;
        TInitFuncList *hooks = getInitHooks();
        for(currentHook  = hooks->begin();
            currentHook != hooks->end() && successful;
            ++currentHook)
            successful = (this->*(currentHook->func))();

        return successful;
    }

protected:
    configurableObject(const char *name) : m_name(name) {}

    // Adds a configurable item to the object (interface for child classes)
    void addConfig(const configItem::EType,
        const void *target,
        const std::string &description,
        const std::vector< std::string > &selections
        = std::vector< std::string >(0));

    // Separate type for configurable functions: Log function
    void addConfigFunc(configItem::EType type,
        configItem::TMemberFunc func,
        const std::string &description = std::string());

    TInitFuncList  m_initHooks;

    virtual TInitFuncList  *getInitHooks()
    {
        return &m_initHooks;
    }

private:
    std::string m_name;
};


/***************************************************************************/
//! The topmost class for all vista-modules
/*!
 *
 * \author  Fabian Wenzel
 *
 ***************************************************************************/
class VISTALABBASEEXPORT module : public configurableObject {
    friend class processCore; // Sets first frame flag
    VISTALABTOPMODULE(module)
public:

    enum EModuleClass{
        INPUTTHREAD,
        INPUT,
        PROCESSING,
        OUTPUT,
        GUI
    };

    // The required interface for *all* modules
    virtual ~module() {};
    virtual bool init()                 =0;
    virtual void done()                 =0;
    virtual bool process()              =0;
    virtual module *clone() const =0;

    // Functions for identification
    const std::string &getInstanceName()   const { return m_instanceName;   }
    const std::string &getHelpString()     const { return m_helpString;     }
    const std::string &getModuleCategory() const { return m_moduleCategory; }
    const std::string &getAuthorInfo()     const { return m_authorEMail;   }

    EModuleClass getModuleClass() const { return m_moduleClass; }

    void setActive(bool active)
    {
        m_active = active;
    }

    void setValid(bool valid)
    {
        m_valid = valid;
    }

    bool isValid()
    {
        return m_valid;
    }

    bool isActive()
    {
        return m_active;
    }

    // Inline - function for transfering input data from the
    // slice to the module (i.e. Input and Outputpins)
    void preProcess( slice &currentSlice )
    {
        TSliceFuncList::iterator currentHook;
        TSliceFuncList *hooks;

        hooks = getPreProcessHooks();
        for(currentHook  = hooks->begin();
            currentHook != hooks->end();
            ++currentHook)
            (this->*(currentHook->func))(currentSlice);
    }


    // Inline - function for transfering output data from the
    // module to the slice
    bool postProcess( slice &currentSlice )
    {
        TPostSliceFuncList::iterator currentHook;
        TPostSliceFuncList *hooks;

        bool result = true;
        hooks = getPostProcessHooks();
        for(currentHook  = hooks->begin();
            currentHook != hooks->end();
            ++currentHook)
            result = result && (this->*(currentHook->func))(currentSlice);
        if(!result)
            setError(_("Did not set all output pins!"));
        return result;
    }


    // The process-wrapper. Calls the virtual process()
    // function of the module after preProcess()ing and
    // before postProcess()ing
    bool process( slice &currentSlice ){
        preProcess(currentSlice);

        if(process()){
            return postProcess(currentSlice);
        } else
            return false;
    }


    // Inline - function which erases the output
    // data of the current slice
    void cleanup( slice &currentSlice ){
        TSliceFuncList *hooks = getCleanupHooks();
        TSliceFuncList::iterator currentHook;
        for(currentHook  = hooks->begin();
            currentHook != hooks->end();
            ++currentHook)
            (this->*(currentHook->func))(currentSlice);
    }

    static int currentFrameNumber()
    {
        return m_processedFramesCounter;
    }

    static void setError(const char *error);
    static const std::string getError();

protected:

    module(EModuleClass moduleClass, const char *moduleType,
        const char *name, const char *helpString, const char *author,
        const char *eMail);

private:
    EModuleClass m_moduleClass;

    // Members for module identification
    PARAMETER(std::string,m_instanceName)
    std::string m_moduleCategory;
    std::string m_helpString;
    std::string m_authorEMail;

    bool        m_active;
    bool        m_valid;

    static int         m_processedFramesCounter;
    static std::string m_error;
};

} // namespace vistaLab

#endif // #define _VLMODULE_H_
